from typing import Optional, List, Any

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils


__all__ = [
    "get_azure_blob_sources",
    "get_single_azure_blob_source",
    "create_azure_blob_source",
]


def get_azure_blob_sources(
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List azure blob sources configured in OIAnalytics.

    Parameters
    ----------
    page : int, default 0
        The page to retrieve. It is ignored if 'get_all_pages' is set to True.
    page_size : int, default 20.
        The size of a page. It is ignored if 'get_all_pages' is set to True.
    get_all_pages : bool, default True
        Whether to get all pages at once.
        If the value is set to True, the 'page' and 'page_size' arguments will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the details of all the azure blob sources, indexed by their IDs.

    """

    if get_all_pages:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        return endpoints.azure_blob_sources.get_azure_blob_sources(
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])
        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "tagContext",
                    "enabled",
                    "name",
                    "pollingRate",
                    "minAge",
                    "maxAge",
                    "regexPattern",
                    "container",
                    "accountName",
                    "customEndpoint",
                    "path",
                    "preserveFiles",
                    "containerNamePrepended",
                    "includeSubdirectories",
                ]
            )
        # Set index
        page_df.set_index("id", inplace=True)
        return page_df

        # Query endpoint

    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_single_azure_blob_source(
    azure_blob_source_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Get a specific azure blob source configured in OIAnalytics by its ID.

    Parameters
    ----------
    azure_blob_source_id : str
        The OIAnalytics ID of the azure blob polling source to retrieve.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas Series object that stores the details of a specific azure blob sourced.

    """
    # get response from the endpoint as a dictionary
    response = endpoints.azure_blob_sources.get_single_azure_blob_source(
        azure_blob_source_id=azure_blob_source_id, api_credentials=api_credentials
    )
    # build a pandas Series with the dictionary
    ser = pd.Series(response, dtype=object)

    # Output
    return ser


def create_azure_blob_source(
    name: str,
    tag_context: List[dict],
    enabled: bool,
    polling_rate: int,
    min_age: str,
    max_age: str,
    regex_pattern: str,
    container: str,
    account_name: Optional[str],
    custom_endpoint: Optional[str],
    account_key: str,
    shared_access_signature: Optional[Any],
    path: str,
    preserve_files: bool,
    container_name_prepended: bool,
    include_subdirectories: bool,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new azure blob source.

    Parameters
    ----------
    name : str
        The name of the azure blob polling source.
    tag_context : List[dict]
        The tag context of the polling source.
    enabled : bool
        Whether the polling of this source is enabled.
    polling_rate : int
        The duration between two polls in seconds.
    min_age : str
        The minimum age a file should have to be considered by a poll.
    max_age : str
        The maximum age a file should have to be considered by a poll.
    regex_pattern : str
        A regex the file name should match to be considered by a poll.
    container : str
        The azure container to poll.
    account_name : str, optional
        The name of the azure account. It must be None if 'custom_endpoint' is not None.
    custom_endpoint : str, optional
        A custom endpoint to use in place of the standard azure endpoint.
        It must be None if 'account_name' is not None.
    account_key : str
        The secret account key attached to the account if the authentication mode requested is by account key.
    shared_access_signature : Any
        The shared access signature to access the resource if the authentication mode requested is by shared access signature.
    path : str
        The path to poll within the container.
    preserve_files : bool
        Whether to preserve files after a pool. If False, then files are deleted.
    container_name_prepended : bool
        If true then the name of the container is added to the file name to form the final name in OIAnalytics.
    include_subdirectories : bool
        If true look for all files also included in sub-directories of the main path that was polled.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the azure blob source that has been created.
    """

    # validation
    if account_name is None and custom_endpoint is None:
        raise ValueError("one of 'account_name' and 'custom_endpoint' should be defined")
    elif account_name is not None and custom_endpoint is not None:
        raise ValueError("one of 'account_name' and 'custom_endpoint' should be None")

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # endpoint
    response = endpoints.azure_blob_sources.create_azure_blob_source(
        name=name,
        tag_context=tag_context,
        enabled=enabled,
        polling_rate=polling_rate,
        min_age=min_age,
        max_age=max_age,
        regex_pattern=regex_pattern,
        container=container,
        account_name=account_name,
        custom_endpoint=custom_endpoint,
        account_key=account_key,
        shared_access_signature=shared_access_signature,
        path=path,
        preserve_files=preserve_files,
        container_name_prepended=container_name_prepended,
        include_subdirectories=include_subdirectories,
        api_credentials=api_credentials,
    )

    ser = pd.Series(response, dtype=object)
    ser.name = ser.loc["name"]

    # Output
    return ser



from typing import Optional

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils


__all__ = [
    "get_tag_keys",
    "get_single_tag_key",
    "get_tag_values",
    "get_single_tag_value",
    "create_tag_value",
    "create_tag_key",
]


def get_tag_keys(
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List all the tag keys, ordered by position.

    Parameters
    ----------
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
        A pandas DataFrame containing tag keys indexed by tag key ID.
    """

    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.tags.get_tag_keys(api_credentials)

    if len(response) > 0:
        df = pd.DataFrame(response)
        df.set_index("id", inplace=True)
    else:
        df = pd.DataFrame(columns=["key", "position", "usedForAccessControl"])
        df.index.name = "id"
        df = df.astype({"key": str, "position": int, "usedForAccessControl": bool})

    # Output
    return df


def get_single_tag_key(
    tag_key_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Retrieve the detail of a specific tag key with its ID.

    Parameters
    ----------
    tag_key_id : str
        The OIAnalytics ID of the tag key.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pandas.Series
        A pandas.Series containing tag key 'id', 'key', 'position' and 'usedForAccessControl'.
    """

    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.tags.get_single_tag_key(
        tag_key_id=tag_key_id, api_credentials=api_credentials
    )

    ser = pd.Series(response, dtype=object)

    # Output
    return ser


def get_tag_values(
    tag_key_id: str,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_tag_key: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List all the tag values of a given tag key.

    Parameters
    ----------
    tag_key_id : str
        The OIAnalytics ID of the tag key.
    expand_tag_key : bool, default True
        Whether to split the column 'tagKey' into multiple columns.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
        A pandas DataFrame, indexed by tag value ID, containing tag values.
    """

    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.tags.get_tag_values(
            tag_key_id=tag_key_id,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "value", "tagKey"])

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    if expand_tag_key:
        df = utils.expand_dataframe_column(
            df=df,
            col="tagKey",
            expected_keys=["id", "key", "position", "usedForAccessControl"],
        )

    # Output
    return df


def get_single_tag_value(
    tag_key_id: str,
    tag_value_id: str,
    expand_tag_key: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Retrieve the detail of a specific tag value with its tag-key ID and tag-value-ID.

    Parameters
    ----------
    tag_key_id : str
        The OIAnalytics ID of the tag key.
    tag_value_id : str
        The OIAnalytics ID of the tag value.
    expand_tag_key : bool
        Whether to split the column 'tagKey' into multiple columns.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas Series containing details of the tag value.
    """
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.tags.get_single_tag_value(
        tag_key_id=tag_key_id,
        tag_value_id=tag_value_id,
        api_credentials=api_credentials,
    )

    df = pd.DataFrame([response])
    if len(df) > 0:
        df.set_index("id", inplace=True)
        if expand_tag_key:
            df = utils.expand_dataframe_column(
                df=df,
                col="tagKey",
            )

    ser = df.squeeze()

    # Output
    return ser


def create_tag_value(
    tag_key_id: str,
    value: str,
    expand_tagkey: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new tag value for the given tag key on the provided command.

    Parameters
    ----------
    tag_key_id : str
        The ID of the tag key of the tag value to be created.
    value : str
        The value of the tag value.
    expand_tagkey : bool, default True
        Whether to split the column 'tagKey'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas.Series with details of the new tag value that has been created.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.tags.create_tag_value(
        tag_key_id=tag_key_id,
        value=value,
        api_credentials=api_credentials,
    )

    df = pd.DataFrame([response])

    if expand_tagkey:
        df = utils.expand_dataframe_column(
            df=df, col="tagKey",
        )

    # Output
    return df.squeeze()


def create_tag_key(
    key: str,
    position: int,
    used_for_access_control: bool,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new tag key on the provided command.

    Parameters
    ----------
    key : str
        The key of the tag key.
    position : int
        The position of the tag key.
    used_for_access_control : str
        Whether this tag key is used to control access objects and data.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the created tag key.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.tags.create_tag_key(
        key=key,
        position=position,
        used_for_access_control=used_for_access_control,
        api_credentials=api_credentials
    )

    ser = pd.Series(response, dtype=object)

    # Output
    return ser



from typing import Union, List, Optional, Dict, Any, Literal
from datetime import datetime, timedelta

import pandas as pd
from pydantic import (
    BaseModel,
    field_validator,
    StrictStr,
    StrictBool,
    StrictInt,
    StrictFloat,
)

from .. import api

__all__ = [
    "ModelExecution",
    "get_default_model_execution",
    "set_default_model_execution",
    "CustomModelOutput",
    "ExecutionReport",
    "get_default_execution_report",
    "set_default_execution_report",
]


class CustomModelOutput(BaseModel):
    type: str
    content: Union[List, StrictStr, Dict]


class ExecutionReport(BaseModel):
    testMode: StrictBool
    status: str = "FAILURE"
    customOutput: Optional[CustomModelOutput] = None
    numberOfTimeValuesCreatedOrUpdated: StrictInt = 0
    numberOfTimeValuesDeleted: StrictInt = 0
    numberOfBatchValuesCreatedOrUpdated: StrictInt = 0
    numberOfBatchTagsCreatedOrUpdated: StrictInt = 0
    numberOfBatchCreatedOrUpdated: StrictInt = 0
    numberOfEventCreatedOrUpdated: StrictInt = 0
    numberOfEventValuesCreatedOrUpdated: StrictInt = 0
    numberOfEventTagsCreatedOrUpdated: StrictInt = 0
    numberOfBatchRelationsCreatedOrUpdated: StrictInt = 0
    numberOfErrors: StrictInt = 0
    numberOfTimeVectorValuesCreatedOrUpdated: StrictInt = 0
    numberOfBatchVectorValuesCreatedOrUpdated: StrictInt = 0
    numberOfFilesUploaded: StrictInt = 0

    def update(
        self,
        time_values_updated: StrictInt = 0,
        time_values_deleted: StrictInt = 0,
        batch_values_updated: StrictInt = 0,
        batch_tags_updated: StrictInt = 0,
        batch_created_updated: StrictInt = 0,
        events_created_updated: StrictInt = 0,
        event_values_updated: StrictInt = 0,
        event_tags_updated: StrictInt = 0,
        batch_relations_updated: StrictInt = 0,
        errors: StrictInt = 0,
        time_vector_values_updated: StrictInt = 0,
        batch_vector_values_updated: StrictInt = 0,
        files_uploaded: StrictInt = 0,
    ):
        self.numberOfTimeValuesCreatedOrUpdated += time_values_updated
        self.numberOfTimeValuesDeleted += time_values_deleted
        self.numberOfBatchValuesCreatedOrUpdated += batch_values_updated
        self.numberOfBatchTagsCreatedOrUpdated += batch_tags_updated
        self.numberOfBatchCreatedOrUpdated += batch_created_updated
        self.numberOfEventCreatedOrUpdated += events_created_updated
        self.numberOfEventValuesCreatedOrUpdated += event_values_updated
        self.numberOfEventTagsCreatedOrUpdated += event_tags_updated
        self.numberOfBatchRelationsCreatedOrUpdated += batch_relations_updated
        self.numberOfErrors += errors
        self.numberOfTimeVectorValuesCreatedOrUpdated += time_vector_values_updated
        self.numberOfBatchVectorValuesCreatedOrUpdated += batch_vector_values_updated
        self.numberOfFilesUploaded += files_uploaded

    def set_as_default_report(self):
        set_default_execution_report(self)


# Credentials
class BasicAuthCredentials(BaseModel):
    baseUrl: StrictStr
    login: StrictStr
    pwd: StrictStr

    def to_object(self):
        return api.OIAnalyticsAPICredentials(
            base_url=self.baseUrl, login=self.login, pwd=self.pwd
        )


class TokenCredentials(BaseModel):
    baseUrl: StrictStr
    token: StrictStr

    def to_object(self):
        return api.OIAnalyticsAPICredentials(base_url=self.baseUrl, token=self.token)


Credentials = Union[BasicAuthCredentials, TokenCredentials]


# Cron Trigger
class CronTrigger(BaseModel):
    type: StrictStr
    cron: StrictStr

    @field_validator("type")
    def check_type(cls, value):
        if value != "cron-trigger":
            raise ValueError("Trigger type should be 'cron-trigger'")
        return value


# File Reception Trigger
class FileReceptionTrigger(BaseModel):
    type: StrictStr

    @field_validator("type")
    def check_type(cls, value):
        if value != "file-reception":
            raise ValueError("Trigger type should be 'file-reception'")
        return value


# Manual only trigger
class ManualOnlyTrigger(BaseModel):
    type: StrictStr

    @field_validator("type")
    def check_type(cls, value):
        if value != "manual-only":
            raise ValueError("Trigger type should be 'manual-only'")
        return value


Trigger = Union[CronTrigger, FileReceptionTrigger, ManualOnlyTrigger]


# Model inputs & outputs
class StringParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[StrictStr] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "STRING":
            raise ValueError("'type' should be 'STRING'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class SecretParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[StrictStr] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "SECRET":
            raise ValueError("'type' should be 'SECRET'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class BooleanParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[StrictBool] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "BOOLEAN":
            raise ValueError("'type' should be 'BOOLEAN'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class NumericParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[Union[StrictInt, StrictFloat]] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "NUMERIC":
            raise ValueError("'type' should be 'NUMERIC'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class FileParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    url: Optional[StrictStr] = None
    value: Optional[StrictStr] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "FILE":
            raise ValueError("'type' should be 'FILE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class PeriodParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[str] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "PERIOD":
            raise ValueError("'type' should be 'PERIOD'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class DurationParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[timedelta] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "DURATION":
            raise ValueError("'type' should be 'DURATION'")
        return value

    @field_validator("value", mode="before")
    def parse_value(cls, value):
        if value is None:
            return None
        else:
            return pd.Timedelta(value, unit="ms").to_pytimedelta()

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class InstantParameter(BaseModel):
    type: Literal["INSTANT"]
    sourceCodeName: StrictStr
    value: Optional[datetime] = None

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class UnitValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class UnitParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[UnitValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "UNIT":
            raise ValueError("'type' should be 'UNIT'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class MeasurementValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class MeasurementParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[MeasurementValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "MEASUREMENT":
            raise ValueError("'type' should be 'MEASUREMENT'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class TagKeyValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class TagKeyParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagKeyValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "TAG_KEY":
            raise ValueError("'type' should be 'TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class TagValueValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class TagValueParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagValueValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "TAG_VALUE":
            raise ValueError("'type' should be 'TAG_VALUE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class DataUnit(BaseModel):
    id: StrictStr
    label: StrictStr


class DataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    unit: DataUnit
    aggregationFunction: Optional[StrictStr] = None


class VectorDataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    valueUnit: DataUnit
    indexUnit: DataUnit
    aggregationFunction: Optional[StrictStr] = None


class DataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[DataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "DATA":
            raise ValueError("'type' should be 'DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class VectorDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[VectorDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "VECTOR_DATA":
            raise ValueError("'type' should be 'VECTOR_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class StoredContinuousDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[DataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "STORED_CONTINUOUS_DATA":
            raise ValueError("'type' should be 'STORED_CONTINUOUS_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchType(BaseModel):
    id: StrictStr
    name: StrictStr


class BatchDataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    unit: DataUnit
    aggregationFunction: Optional[StrictStr] = None
    batchType: BatchType


class EventDataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    unit: DataUnit
    aggregationFunction: Optional[StrictStr] = None


class BatchVectorDataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    valueUnit: DataUnit
    indexUnit: DataUnit
    aggregationFunction: Optional[StrictStr] = None
    indexAggregationFunction: Optional[StrictStr] = None
    batchType: BatchType
    batchesIds: Optional[List[StrictStr]] = None


class StoredBatchDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "STORED_BATCH_DATA":
            raise ValueError("'type' should be 'STORED_BATCH_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class StoredEventDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "STORED_EVENT_DATA":
            raise ValueError("'type' should be 'STORED_EVENT_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class StoredBatchVectorDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchVectorDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "STORED_BATCH_VECTOR_DATA":
            raise ValueError("'type' should be 'STORED_BATCH_VECTOR_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class StoredTimeVectorDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[VectorDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "STORED_CONTINUOUS_VECTOR_DATA":
            raise ValueError("'type' should be 'STORED_CONTINUOUS_VECTOR_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class ComputedContinuousDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[DataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "COMPUTED_CONTINUOUS_DATA":
            raise ValueError("'type' should be 'COMPUTED_CONTINUOUS_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class ComputedBatchDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "COMPUTED_BATCH_DATA":
            raise ValueError("'type' should be 'COMPUTED_BATCH_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchTimeDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchDataValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "BATCH_TIME_DATA":
            raise ValueError("'type' should be 'BATCH_TIME_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchPredicateBatchType(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class BatchTagValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr

    @field_validator("type")
    def check_type(cls, value):
        if value != "batch-tag-value":
            raise ValueError("Batch tag value type should be 'batch-tag-value'")
        return value


class BatchPredicateValue(BaseModel):
    type: StrictStr
    batchType: BatchPredicateBatchType
    featureFilters: List[BatchTagValue]

    @field_validator("type")
    def check_type(cls, value):
        if value != "batch-predicate":
            raise ValueError("Batch predicate type should be 'batch-predicate'")
        return value


class BatchPredicateParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchPredicateValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "BATCH_PREDICATE":
            raise ValueError("'type' should be 'BATCH_PREDICATE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.batchType.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.batchType.id


class BatchStructureValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class BatchStructureParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchStructureValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "BATCH_STRUCTURE":
            raise ValueError("'type' should be 'BATCH_STRUCTURE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchTagKeyParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagKeyValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "BATCH_TAG_KEY":
            raise ValueError("'type' should be 'BATCH_TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchTagValueParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagValueValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "BATCH_TAG_VALUE":
            raise ValueError("'type' should be 'BATCH_TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class EventTypeValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class EventTypeParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventTypeValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "EVENT_TYPE":
            raise ValueError("'type' should be 'EVENT_TYPE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class EventTagKeyValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class EventTagKeyParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventTagKeyValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "EVENT_TAG_KEY":
            raise ValueError("'type' should be 'EVENT_TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class EventTagValueValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class EventTagValueParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventTagValueValue] = None

    @field_validator("type")
    def validate_type(cls, value):
        if value != "EVENT_TAG_VALUE":
            raise ValueError("'type' should be 'EVENT_TAG_VALUE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class TimeDataParametersListValue(BaseModel):
    type: Literal["temporal-data-list"]
    dataList: List[DataValue]


class TimeDataParametersList(BaseModel):
    type: Literal["TEMPORAL_DATA_LIST"]
    sourceCodeName: StrictStr
    value: TimeDataParametersListValue

    @property
    def readable_value(self) -> List[StrictStr]:
        return [value.reference for value in self.value.dataList]

    @property
    def technical_value(self) -> List[StrictStr]:
        return [value.id for value in self.value.dataList]


class BatchDataParametersListValue(BaseModel):
    type: Literal["batch-data-list"]
    dataList: List[BatchDataValue]


class BatchDataParametersList(BaseModel):
    type: Literal["BATCH_DATA_LIST"]
    sourceCodeName: StrictStr
    value: BatchDataParametersListValue

    @property
    def readable_value(self) -> List[StrictStr]:
        return [value.reference for value in self.value.dataList]

    @property
    def technical_value(self) -> List[StrictStr]:
        return [value.id for value in self.value.dataList]


InstanceParameter = Union[
    StringParameter,
    SecretParameter,
    BooleanParameter,
    NumericParameter,
    FileParameter,
    PeriodParameter,
    DurationParameter,
    InstantParameter,
    UnitParameter,
    MeasurementParameter,
    TagKeyParameter,
    TagValueParameter,
    DataParameter,
    TimeDataParametersList,
    BatchDataParametersList,
    StoredContinuousDataParameter,
    StoredBatchDataParameter,
    StoredTimeVectorDataParameter,
    StoredBatchVectorDataParameter,
    ComputedContinuousDataParameter,
    ComputedBatchDataParameter,
    BatchTimeDataParameter,
    BatchPredicateParameter,
    BatchStructureParameter,
    BatchTagKeyParameter,
    BatchTagValueParameter,
    StoredEventDataParameter,
    EventTypeParameter,
    EventTagKeyParameter,
    EventTagValueParameter,
]


# Model instance
class SingleObservationTimeContext(BaseModel):
    type: StrictStr
    overlappingPeriod: timedelta
    aggregationPeriod: StrictStr

    @field_validator("type")
    def check_type(cls, value):
        if value != "time":
            raise ValueError("Single observation type should be 'time'")
        return value

    @field_validator("overlappingPeriod", mode="before")
    def parse_overlapping_period(cls, value):
        return pd.Timedelta(value).to_pytimedelta()


class SingleObservationBatchContext(BaseModel):
    type: StrictStr
    batchPredicate: BatchPredicateValue
    overlappingPeriod: timedelta

    @field_validator("type")
    def check_type(cls, value):
        if value != "batch":
            raise ValueError("Single observation type should be 'batch'")
        return value

    @field_validator("overlappingPeriod", mode="before")
    def parse_overlapping_period(cls, value):
        return pd.Timedelta(value).to_pytimedelta()


SingleObservationContext = Union[
    SingleObservationBatchContext, SingleObservationTimeContext
]


class ExecutionContext(BaseModel):
    type: StrictStr
    initiator: StrictStr
    lastSuccessfulExecutionInstant: Optional[datetime] = None
    currentExecutionInstant: datetime
    customData: Optional[Any] = None
    uploadEventId: Optional[StrictStr] = None
    fileName: Optional[StrictStr] = None


class ModelInstance(BaseModel):
    id: StrictStr
    trigger: Trigger
    active: StrictBool
    dataExchangeMode: StrictStr
    singleObservationContext: Optional[SingleObservationContext] = None
    inputParameters: List[InstanceParameter]
    outputParameters: List[InstanceParameter]


# Model execution
class ModelExecution(BaseModel):
    testMode: StrictBool
    credentials: Optional[Credentials] = None
    lastSuccessfulExecutionInstant: Optional[datetime] = None
    currentExecutionInstant: datetime
    executionContext: ExecutionContext
    pythonModelInstance: ModelInstance

    def set_as_default_model_execution(self):
        set_default_model_execution(self)

    def find_inputs_by_types(self, input_types: Optional[List[str]] = None):
        if input_types is None:
            return self.pythonModelInstance.inputParameters
        else:
            return [
                param
                for param in self.pythonModelInstance.inputParameters
                if param.type in input_types
            ]

    def find_input_by_sourcecodename(self, source_code_name: str):
        for param in self.pythonModelInstance.inputParameters:
            if param.sourceCodeName == source_code_name:
                return param
            raise KeyError(
                f"No input parameter is referenced by the source code name {source_code_name}"
            )

    def find_outputs_by_types(self, output_types: Optional[List[str]] = None):
        if output_types is None:
            return self.pythonModelInstance.outputParameters
        else:
            return [
                param
                for param in self.pythonModelInstance.outputParameters
                if param.type in output_types
            ]

    def find_output_by_sourcecodename(self, source_code_name: str):
        for param in self.pythonModelInstance.outputParameters:
            if param.sourceCodeName == source_code_name:
                return param
            raise KeyError(
                f"No output parameter is referenced by the source code name {source_code_name}"
            )

    def get_input_dict(
        self,
        input_types: Optional[List[str]] = None,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        input_params = self.find_inputs_by_types(input_types)
        if mode == "reference":
            input_dict = {
                param.sourceCodeName: param.readable_value for param in input_params
            }
        elif mode == "id":
            input_dict = {
                param.sourceCodeName: param.technical_value for param in input_params
            }
        elif mode == "object":
            input_dict = {param.sourceCodeName: param.value for param in input_params}
        else:
            raise ValueError(f"Unknown 'mode': {mode}")

        if exclude_empty_values is True:
            input_dict = {k: v for k, v in input_dict.items() if v is not None}

        return input_dict

    def get_output_dict(
        self,
        output_types: Optional[List[str]] = None,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        output_params = self.find_outputs_by_types(output_types)
        if mode == "reference":
            output_dict = {
                param.sourceCodeName: param.readable_value for param in output_params
            }
        elif mode == "id":
            output_dict = {
                param.sourceCodeName: param.technical_value for param in output_params
            }
        elif mode == "object":
            output_dict = {param.sourceCodeName: param.value for param in output_params}
        else:
            raise ValueError(f"Unknown value for 'mode': {mode}")

        if exclude_empty_values is True:
            output_dict = {k: v for k, v in output_dict.items() if v is not None}

            return output_dict

    def get_data_input_dict(
        self,
        data_type: str = "any",
        values_type: str = "any",
        strict: bool = False,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        """
        Parameters
        ----------
        data_type: {'any, 'time', 'batch'}
        values_type: {'any', 'scalar', 'vector'}
        strict: bool
        mode: {'object', 'reference', 'id'}
        exclude_empty_values: bool
        """
        if data_type == "any" and values_type == "any":
            input_types = [
                "DATA",
                "STORED_CONTINUOUS_DATA",
                "STORED_BATCH_DATA",
                "COMPUTED_CONTINUOUS_DATA",
                "TEMPORAL_DATA_LIST",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_DATA_LIST",
                "BATCH_TAG_KEY",
                "STORED_CONTINUOUS_VECTOR_DATA",
                "STORED_BATCH_VECTOR_DATA",
            ]
        elif data_type == "any" and values_type == "scalar":
            input_types = [
                "DATA",
                "STORED_CONTINUOUS_DATA",
                "STORED_BATCH_DATA",
                "COMPUTED_CONTINUOUS_DATA",
                "TEMPORAL_DATA_LIST",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_DATA_LIST",
                "BATCH_TAG_KEY",
            ]
        elif data_type == "any" and values_type == "vector":
            input_types = [
                "STORED_CONTINUOUS_VECTOR_DATA",
                "STORED_BATCH_VECTOR_DATA",
            ]
        elif data_type == "time" and values_type == "any":
            input_types = [
                "STORED_CONTINUOUS_DATA",
                "COMPUTED_CONTINUOUS_DATA",
                "STORED_CONTINUOUS_VECTOR_DATA",
                "TEMPORAL_DATA_LIST",
            ]
            if strict is False:
                input_types.append("DATA")
        elif data_type == "time" and values_type == "scalar":
            input_types = [
                "STORED_CONTINUOUS_DATA",
                "COMPUTED_CONTINUOUS_DATA",
                "TEMPORAL_DATA_LIST",
            ]
            if strict is False:
                input_types.append("DATA")
        elif data_type == "time" and values_type == "vector":
            input_types = ["STORED_CONTINUOUS_VECTOR_DATA"]
            if strict is False:
                input_types.append("VECTOR_DATA")
        elif data_type == "batch" and values_type == "any":
            input_types = [
                "STORED_BATCH_DATA",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_DATA_LIST",
                "BATCH_TAG_KEY",
                "STORED_BATCH_VECTOR_DATA",
            ]
            if strict is False:
                input_types.append("DATA")
        elif data_type == "batch" and values_type == "scalar":
            input_types = [
                "STORED_BATCH_DATA",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_DATA_LIST",
                "BATCH_TAG_KEY",
            ]
            if strict is False:
                input_types.append("DATA")
        elif data_type == "batch" and values_type == "vector":
            input_types = ["STORED_BATCH_VECTOR_DATA"]
            if strict is False:
                input_types.append("VECTOR_DATA")
        else:
            raise ValueError(f"Unknown 'data_type': {data_type}")

        return self.get_input_dict(
            input_types=input_types,
            mode=mode,
            exclude_empty_values=exclude_empty_values,
        )

    def get_data_output_dict(
        self,
        data_type: str = "any",
        values_type: str = "any",
        strict: bool = False,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        """
        Parameters
        ----------
        data_type: {'any, 'time', 'batch', 'event'}
        values_type: {'any', 'scalar', 'categorical', 'vector'}
        strict: bool
        mode: {'object', 'reference', 'id'}
        exclude_empty_values: bool
        """
        if data_type == "any":
            if values_type == "any":
                output_types = [
                    "DATA",
                    "STORED_CONTINUOUS_DATA",
                    "STORED_BATCH_DATA",
                    "COMPUTED_CONTINUOUS_DATA",
                    "COMPUTED_BATCH_DATA",
                    "BATCH_TIME_DATA",
                    "BATCH_TAG_KEY",
                    "STORED_BATCH_VECTOR_DATA",
                    "STORED_CONTINUOUS_VECTOR_DATA",
                    "STORED_EVENT_DATA",
                    "EVENT_TAG_KEY",
                ]
            elif values_type == "scalar":
                output_types = [
                    "DATA",
                    "STORED_CONTINUOUS_DATA",
                    "STORED_BATCH_DATA",
                    "COMPUTED_CONTINUOUS_DATA",
                    "COMPUTED_BATCH_DATA",
                    "BATCH_TIME_DATA",
                    "STORED_EVENT_DATA",
                ]
            elif values_type == "categorical":
                output_types = [
                    "BATCH_TAG_KEY",
                    "EVENT_TAG_KEY",
                ]
            elif values_type == "vector":
                output_types = [
                    "STORED_CONTINUOUS_VECTOR_DATA",
                    "STORED_BATCH_VECTOR_DATA",
                ]

        elif data_type == "time":
            if values_type == "any":
                output_types = [
                    "STORED_CONTINUOUS_DATA",
                    "COMPUTED_CONTINUOUS_DATA",
                    "STORED_CONTINUOUS_VECTOR_DATA",
                ]
                if strict is False:
                    output_types.append("DATA")
            elif values_type == "scalar":
                output_types = ["STORED_CONTINUOUS_DATA", "COMPUTED_CONTINUOUS_DATA"]
                if strict is False:
                    output_types.append("DATA")
            elif values_type == "vector":
                output_types = ["STORED_CONTINUOUS_VECTOR_DATA"]
                if strict is False:
                    output_types.append("VECTOR_DATA")

        elif data_type == "batch":
            if values_type == "any":
                output_types = [
                    "STORED_BATCH_DATA",
                    "COMPUTED_BATCH_DATA",
                    "BATCH_TIME_DATA",
                    "BATCH_TAG_KEY",
                    "STORED_BATCH_VECTOR_DATA",
                ]
                if strict is False:
                    output_types.append("DATA")
            elif values_type == "scalar":
                output_types = [
                    "STORED_BATCH_DATA",
                    "COMPUTED_BATCH_DATA",
                    "BATCH_TIME_DATA",
                ]
                if strict is False:
                    output_types.append("DATA")
            elif values_type == "categorical":
                output_types = [
                    "BATCH_TAG_KEY",
                ]
            elif values_type == "vector":
                output_types = ["STORED_BATCH_VECTOR_DATA"]
                if strict is False:
                    output_types.append("VECTOR_DATA")

        elif data_type == "event":
            if values_type == "any":
                output_types = ["STORED_EVENT_DATA", "EVENT_TAG_KEY"]
                if strict is False:
                    output_types.append("DATA")
            elif values_type == "scalar":
                output_types = ["STORED_EVENT_DATA"]
                if strict is False:
                    output_types.append("DATA")
            elif values_type == "categorical":
                output_types = ["EVENT_TAG_KEY"]

        else:
            raise ValueError(f"Unknown 'data_type': {data_type}")

        return self.get_output_dict(
            output_types=output_types,
            mode=mode,
            exclude_empty_values=exclude_empty_values,
        )

    @property
    def execution_kwargs(self):
        return {
            "model_instance_id": self.pythonModelInstance.id,
            "test_mode": self.testMode,
            "execution_context": self.executionContext.dict(),
            "current_execution_date": self.currentExecutionInstant,
            "last_execution_date": self.lastSuccessfulExecutionInstant,
            "input_parameters": self.get_input_dict(mode="object"),
            "input_parameter_references": self.get_input_dict(mode="reference"),
            "input_parameter_ids": self.get_input_dict(mode="id"),
            "output_parameters": self.get_output_dict(mode="object"),
            "output_parameter_references": self.get_output_dict(mode="reference"),
            "output_parameter_ids": self.get_output_dict(mode="id"),
            "custom_input_data": self.executionContext.customData,
        }

    def get_execution_argument_value(self, argument_name: str):
        return self.execution_kwargs[argument_name]


# Current event
DEFAULT_EVENT: Optional[ModelExecution] = None


def set_default_model_execution(event: ModelExecution):
    """
    Put an execution event into a global variable, for external use

    Parameters
    ----------
    event: ModelExecution
    """

    global DEFAULT_EVENT
    DEFAULT_EVENT = event


def get_default_model_execution():
    global DEFAULT_EVENT
    return DEFAULT_EVENT


# Current summary
DEFAULT_EXECUTION_REPORT: Optional[ExecutionReport] = None


def set_default_execution_report(execution_report: ExecutionReport):
    """
    Put an execution report into a global variable, for external use

    Parameters
    ----------
    execution_report: ExecutionReport
    """

    global DEFAULT_EXECUTION_REPORT
    DEFAULT_EXECUTION_REPORT = execution_report


def get_default_execution_report():
    """
    Get the current default execution report.
    If none has been set yet, it creates an empty one.

    Returns
    -------
    The current default execution report
    """

    global DEFAULT_EXECUTION_REPORT
    if DEFAULT_EXECUTION_REPORT is None:
        DEFAULT_EXECUTION_REPORT = ExecutionReport(testMode=True)
    return DEFAULT_EXECUTION_REPORT



from typing import Optional, Union, List
from datetime import datetime

import requests

from .. import _credentials
from .. import utils


__all__ = [
    "get_event_types",
    "get_event_type_details",
    "create_event_type",
    "update_event_type",
    "get_single_event",
    "get_events",
    "delete_event",
]


def get_event_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/event-types"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_event_type_details(
    event_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/event-types/{event_type_id}"
    response = requests.get(url=url, **api_credentials.auth_kwargs)

    # Output
    response.raise_for_status()
    return response.json()


def get_single_event(
    event_type_id: str,
    event_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/event-types/{event_type_id}/events/{event_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def create_event_type(
    name: str,
    asset_type_ids: List[str],
    tag_keys: List[dict],
    tag_context: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/event-types"
    response = requests.post(
        url=url,
        json={
            "name": name,
            "assetTypeIds": asset_type_ids,
            "tagKeys": tag_keys,
            "tagContext": tag_context,
        },
        **api_credentials.auth_kwargs
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_event_type(
    event_type_id: str,
    name: str,
    asset_type_ids: List[str],
    tag_keys: List[dict],
    tag_context: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing event type.

    Parameters
    ----------
    event_type_id : str
        The OIAnalytics ID of the event type to update.
    name : str
        The name of the event type to create.
    asset_type_ids : list of str
        The IDs of the types of asset that can fire this type of event.
    tag_keys : list of dict
        The array of source tag keys to create on the event type.
    tag_context : list of dict
        The array of tag access to attach to the event type to restrict access of users.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/event-types/{event_type_id}"
    response = requests.put(
        url=url,
        json={
            "name": name,
            "assetTypeIds": asset_type_ids,
            "tagKeys": tag_keys,
            "tagContext": tag_context,
        },
        **api_credentials.auth_kwargs
    )

    # Output
    response.raise_for_status()
    return response.status_code


def get_events(
    event_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    description: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Query endpoint
    url = (
        f"{api_credentials.base_url}/api/oianalytics/event-types/{event_type_id}/events"
    )
    response = requests.get(
        url=url,
        params={
            "start": start_date_iso,
            "end": end_date_iso,
            "description": description,
            "feature-values": features_value_ids,
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_event(
    event_type_id: str,
    event_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/event-types/{event_type_id}/events/{event_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def create_or_update_events(
    event_type_id: str,
    events: List[dict],
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    url = f"{api_credentials.base_url}/api/oianalytics/event-types/{event_type_id}/events/create-or-update"
    response = requests.post(
        url=url,
        json=events,
        params={"create-upload-event": create_upload_event},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Optional, Union
import requests
import urllib3
import io

from ... import api
from .. import _dtos

__all__ = ["get_resource_file", "update_instance_resource"]


def get_resource_file(
    resource_file_id: str,
    model_instance_id: Optional[str] = None,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = api.get_default_oianalytics_credentials()

    # Get model execution from environment if not provided
    if model_instance_id is None:
        model_instance_id = _dtos.get_default_model_execution().pythonModelInstance.id

    if model_instance_id == _dtos.get_default_model_execution().pythonModelInstance.id:
        resource_files = _dtos.get_default_model_execution().find_inputs_by_types(
            "FILE"
        )
        for rf in resource_files:
            if rf.value == resource_file_id:
                resource_file_url = rf.url
                if resource_file_url is not None:
                    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
                    return io.BytesIO(
                        requests.get(resource_file_url, verify=False).content
                    )

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/python-model-instances/{model_instance_id}/resources/{resource_file_id}/content"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return io.BytesIO(response.content)


def update_instance_resource(
    file_content: Union[io.StringIO, io.BytesIO],
    file_name: str,
    resource_file_id: str,
    model_instance_id: Optional[str] = None,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = api.get_default_oianalytics_credentials()

    # Get model execution from environment if not provided
    if model_instance_id is None:
        model_instance_id = _dtos.get_default_model_execution().pythonModelInstance.id

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/python-model-instances/{model_instance_id}/resources/{resource_file_id}/content"
    response = requests.post(
        url=url,
        files={"file": (file_name, file_content)},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code



from .batches import (
    get_batch_types,
    get_batch_type_details,
    get_single_batch,
    get_batches,
    update_batch,
    update_batch_values,
    update_batch_feature_values,
    update_batch_features_and_values,
    update_vector_batch_values,
    create_or_update_batches,
    get_batch_relations,
    get_single_batch_relation,
)
from .azure_blob_sources import (
    get_azure_blob_sources,
    get_single_azure_blob_source,
    create_azure_blob_source,
)
from .data import (
    get_data_list,
    get_time_values,
    get_vector_time_values,
    get_batch_values,
    get_vector_batch_values,
    get_multiple_data_values,
    insert_time_values,
    insert_vector_time_values,
)
from .tags import (
    get_tag_keys,
    get_single_tag_key,
    get_tag_values,
    get_single_tag_value,
    create_tag_key,
    create_tag_value,
)
from .events import (
    get_event_types,
    get_events,
    create_or_update_events,
    get_event_type_details,
    get_events,
    create_event_type,
)
from .dashboard import get_dashboards
from .files import get_file_uploads, read_file_from_file_upload
from .users import get_users, get_user_details, create_user
from .profiles import (
    get_permissions,
    get_profiles,
    get_single_profile,
    create_profile,
)
from .assets import (
    get_asset_types,
    get_single_asset_type,
    create_asset_type,
    get_assets,
    get_single_asset,
    update_single_asset_tags_and_values,
    create_asset,
)
from .quantities import get_quantities
from .units import get_units, get_unit_families
from .model_instances import get_model_instances, get_single_model_instance

__all__ = [
    "get_batch_types",
    "get_batch_type_details",
    "get_single_batch",
    "get_batches",
    "update_batch",
    "update_batch_values",
    "update_batch_feature_values",
    "update_batch_features_and_values",
    "get_batch_relations",
    "get_single_batch_relation",
    "get_data_list",
    "get_time_values",
    "get_vector_time_values",
    "get_batch_values",
    "get_vector_batch_values",
    "get_multiple_data_values",
    "insert_time_values",
    "insert_vector_time_values",
    "update_vector_batch_values",
    "create_or_update_batches",
    "get_azure_blob_sources",
    "get_single_azure_blob_source",
    "create_azure_blob_source",
    "get_event_types",
    "get_event_type_details",
    "create_event_type",
    "get_events",
    "create_or_update_events",
    "get_dashboards",
    "get_file_uploads",
    "read_file_from_file_upload",
    "get_users",
    "get_user_details",
    "create_user",
    "get_permissions",
    "get_profiles",
    "get_single_profile",
    "create_profile",
    "get_asset_types",
    "get_single_asset_type",
    "create_asset_type",
    "get_assets",
    "get_single_asset",
    "update_single_asset_tags_and_values",
    "create_asset",
    "get_quantities",
    "get_units",
    "get_unit_families",
    "get_tag_keys",
    "get_single_tag_key",
    "get_tag_values",
    "create_tag_key",
    "get_single_tag_value",
    "create_tag_value",
    "get_model_instances",
    "get_single_model_instance",
]



from typing import Optional
import requests

from .. import _credentials


__all__ = [
    "get_model_instances",
    "get_single_model_instance",
]


def get_model_instances(
    page: int = 0,
    page_size: int = 20,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/python-model-instances"
    response = requests.get(
        url=url,
        params={"page": page, "size": page_size},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_model_instance(
    model_instance_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/python-model-instances/{model_instance_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Optional, List
import requests

from .. import _credentials


__all__ = [
    "get_permissions",
    "get_profiles",
    "get_single_profile",
    "delete_profile",
    "create_profile",
    "update_profile",
]


def get_permissions(
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/profiles/permissions"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_profiles(
    page: int = 0,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/profiles"
    response = requests.get(
        url=url,
        params={"page": page},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_profile(
    profile_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/profiles/{profile_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_profile(
    profile_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an existing profile with a given id.

    Parameters
    ----------
    profile_id : str
        The OIAnalytics ID of the profile to retrieve.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/profiles/{profile_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    response.raise_for_status()
    return response.status_code


def create_profile(
    name: str,
    external_id: Optional[str],
    access_rights: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/profiles"
    response = requests.post(
        url=url,
        json={
            "name": name,
            "externalId": external_id,
            "accessRights": access_rights,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_profile(
    profile_id: str,
    name: str,
    external_id: Optional[str],
    access_rights: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing user.

    Parameters
    ----------
    profile_id : str
        The OIAnalytics ID of the profile to update.
    name : str
        The name of the profile.
    external_id : str
        The ID of a group in identity provider to associate with the profile.
    access_rights : list of dict
        The access rights associated with this profile.
        Each dictionary in the list contains the following keys:
        - 'permission': str, the permission the access right defines.
        - 'mode': {'YES', 'NO','READ',WRITE'}, the mode of access right is permitted.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/profiles/{profile_id}"
    response = requests.put(
        url=url,
        json={
            "name": name,
            "externalId": external_id,
            "accessRights": access_rights,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code



from . import users
from . import profiles
from . import data
from . import files
from . import azure_blob_sources
from . import batches
from . import events
from . import dashboard
from . import assets
from . import quantities
from . import units
from . import tags
from . import model_instances
from . import computation_jobs

from .users import delete_user, update_user
from .profiles import delete_profile, update_profile
from .azure_blob_sources import delete_azure_blob_source, update_azure_blob_source
from .batches import delete_batch
from .events import update_event_type
from .tags import delete_single_tag_key, delete_single_tag_value, update_tag_key, update_tag_value
from .assets import delete_asset, update_asset, delete_asset_type, update_asset_type

__all__ = [
    "update_user",
    "delete_user",
    "delete_profile",
    "update_azure_blob_source",
    "update_profile",
    "delete_azure_blob_source",
    "delete_batch",
    "update_event_type",
    "update_asset_type",
    "delete_asset_type",
    "update_asset",
    "delete_asset",
    "delete_single_tag_key",
    "delete_single_tag_value",
    "update_tag_value",
    "update_tag_key",
]



import importlib.resources as pkg_resources

from . import _template_resources


# List templates
def available_templates():
    return ["free_from_api", "single_observation", "file_processing"]


# Show templates
def get_template(template: str):
    """
    Read the source code of a template file

    Parameters
    ----------
    template: str
        Name of the template. Can be 'free_from_api', 'single_observation' or 'file_processing'.

    Returns
    -------
    The source code string of the template
    """

    return pkg_resources.read_text(_template_resources, f"{template}_template.py")


def print_template(template: str):
    """
    Print the source code of a template file

    Parameters
    ----------
    template: str
        Name of the template. Can be 'free_from_api', 'single_observation' or 'file_processing'.

    Returns
    -------
    Doesn't return outputs. Prints the source code string of the template.
    """

    print(get_template(template=template))



from typing import Optional
import requests

from ... import api

__all__ = ["invoke_instance"]


def invoke_instance(
    model_instance_id: str,
    custom_input_data=None,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = api.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/python-model-instances/{model_instance_id}/invoke"
    response = requests.post(
        url=url,
        json=custom_input_data,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



import requests
from typing import Optional, Literal

from .. import _credentials


__all__ = ["get_dashboards", "get_single_dashboard"]


def get_dashboards(
    page: int = 0,
    page_size: int = 20,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/dashboard"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_dashboard(
    dashboard_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/dashboard/{dashboard_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Optional

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils


__all__ = ["get_quantities"]


def get_quantities(
    name: Optional[str] = None,
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    List physical quantities by page.

    Parameters
    ----------
    name : str, optional
        The text the quantity should contain. The search is performed in case-insensitive manner.
    page : int, default 0
        The page to retrieve. It is ignored if 'get_all_pages' is set to True.
    page_size : int, default 20.
        The size of one page. It is ignored if 'get_all_pages' is set to True.
    get_all_pages : bool = True
        Whether to get quantities of all pages.
        If the value is set to True, the 'page' and 'page_size' arguments will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
        A pandas Dataframe containing the quantities names indexed by their corresponding quantities ids.
    """

    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int) -> dict:
        return endpoints.quantities.get_quantities(
            name=name,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )

    def parse_page(page_response: dict) -> pd.DataFrame:
        page_df = pd.DataFrame(page_response["content"], dtype=str)
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "name",
                ]
            )
        page_df.set_index("id", inplace=True)
        return page_df

    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df.squeeze()



import pandas as pd
from typing import Optional

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = ["get_dashboards"]


def get_dashboards(
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List azure blob sources configured in OIAnalytics.

    Parameters
    ----------
    page : int
        The page to retrieve, defaults to 0 (first page).
    page_size : int
        The size of a page.
    get_all_pages : bool, default True
        Whether to list all pages of dashboards.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame with all the dashboards.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    if get_all_pages is True:
        page = 0

    def get_page(page_num: int):
        page_response = endpoints.dashboard.get_dashboards(
            page=page_num, page_size=page_size, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "title", "description", "type"])

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df



from typing import Optional, Dict, List, Literal, Any

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = [
    "get_asset_types",
    "get_single_asset_type",
    "create_asset_type",
    "get_assets",
    "get_single_asset",
    "update_single_asset_tags_and_values",
    "create_asset",
]


def get_asset_types(
    page: Optional[int] = None,
    name: Optional[str] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    extract_from_tagcontext: Optional[str] = "value",
    expand_tagcontext: bool = True,
    extract_from_dataset: Optional[str] = "description",
    extract_from_staticdataset: Optional[str] = "description",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List asset types by page according to the provided filter criteria.

    Parameters
    ----------
    page : int, optional
        The page to retrieve, defaults to first page.
    name : int, optional
        A string that should be contained by all asset type names returned.
    get_all_pages : bool, default True
        Whether to get asset types of all pages.
        If the value is set to True, the 'page' argument will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    extract_from_tagcontext : str, optional, default 'value'
        The attribute to extract from the column 'tagContext'.
    expand_tagcontext : bool, default True
        Whether to expand the column 'tagContext'.
    extract_from_dataset : str, optional, default 'description'
        The attribute to extract from the column 'dataSet'.
    extract_from_staticdataset : str, optional, default 'description'
        The attribute to extract from the column 'staticDataSet'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing details of the requested asset types.
    """
    # Args validation
    if extract_from_tagcontext not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_tagcontext': {extract_from_tagcontext}"
        )

    if extract_from_tagcontext is None and expand_tagcontext is True:
        raise ValueError(
            "Tag context cannot be expanded if 'extract_from_tagcontext' is None"
        )

    if extract_from_dataset not in ["id", "name", "description", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_dataset': {extract_from_dataset}"
        )

    if extract_from_staticdataset not in ["id", "name", "description", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_staticdataset': {extract_from_staticdataset}"
        )

    # Init
    if get_all_pages is True:
        page = 0

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    def get_page(page_num: int):
        page_response = endpoints.assets.get_asset_types(
            page=page_num, name=name, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "name"])

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Format dataframe
    if extract_from_tagcontext == "value":
        df["tagContext"] = df["tagContext"].map(
            lambda tagcontext: {
                tag["tagKey"]["key"]: [
                    val["value"] for val in tag["accessibleTagValues"]
                ]
                for tag in tagcontext
            }
        )

    elif extract_from_tagcontext == "id":
        df["tagContext"] = df["tagContext"].map(
            lambda tagcontext: {
                tag["tagKey"]["id"]: [val["id"] for val in tag["accessibleTagValues"]]
                for tag in tagcontext
            }
        )

    if expand_tagcontext is True and extract_from_tagcontext is not None:
        df = utils.expand_dataframe_column(df, "tagContext", add_prefix=False)

    if extract_from_dataset is not None:
        df["numericDataSet"] = df["numericDataSet"].map(
            lambda dataset: [data[extract_from_dataset] for data in dataset]
        )
        df["stringDataSet"] = df["stringDataSet"].map(
            lambda dataset: [data[extract_from_dataset] for data in dataset]
        )

    if extract_from_staticdataset is not None:
        df["staticDataSet"] = df["staticDataSet"].map(
            lambda staticdataset: [
                data[extract_from_staticdataset] for data in staticdataset
            ]
        )

    # Output
    return df


def get_single_asset_type(
    asset_type_id: str,
    extract_from_dataset: Optional[
        Literal["id", "name", "description"]
    ] = "description",
    extract_from_tagcontext: Optional[Literal["id", "value"]] = "value",
    expand_tagcontext: bool = True,
    extract_from_staticdataset: Optional[
        Literal["id", "name", "description"]
    ] = "description",
    extract_from_pythonmodels: Optional[
        Literal["id", "name", "description"]
    ] = "description",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Retrieve the details of a specific asset type with its ID.

    Parameters
    ----------
    asset_type_id : str
        The ID of the asset type to retrieve.
    extract_from_dataset : {'id', 'name', 'description'}, optional, default 'description'
        Key to be associated to the column 'dataSet'.
    extract_from_tagcontext : {'id', 'value'}, optional, default 'value'
        Key to be associated from the column 'tagContext'.
    expand_tagcontext : bool, default True
        Whether to expand the column 'tagContext'.
    extract_from_staticdataset : {'id', 'name', 'description}, optional, default 'description'
        Key to be associated to the column 'staticDataSet'.
    extract_from_pythonmodels : {'id', 'name', 'description}, optional, default 'description'
        Key to be associated to the column 'pythonModels'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas Series containing the details of a specific asset type.
    """

    # Args validation
    if extract_from_tagcontext is None and expand_tagcontext is True:
        raise ValueError(
            "Tag context cannot be expanded if 'extract_from_tagcontext' is None"
        )

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.assets.get_single_asset_type(
        asset_type_id=asset_type_id, api_credentials=api_credentials
    )

    df = pd.DataFrame([response])
    df.set_index("id", inplace=True)

    # Format DataFrame
    if extract_from_dataset:
        df["numericDataSet"] = df["numericDataSet"].map(
            lambda dataset: [data[extract_from_dataset] for data in dataset]
        )
        df["stringDataSet"] = df["stringDataSet"].map(
            lambda dataset: [data[extract_from_dataset] for data in dataset]
        )

    if extract_from_tagcontext == "value":
        df["tagContext"] = df["tagContext"].map(
            lambda tagcontext: {
                tag["tagKey"]["key"]: [
                    val["value"] for val in tag["accessibleTagValues"]
                ]
                for tag in tagcontext
            }
        )
    elif extract_from_tagcontext == "id":
        df["tagContext"] = df["tagContext"].map(
            lambda tagcontext: {
                tag["tagKey"]["id"]: [val["id"] for val in tag["accessibleTagValues"]]
                for tag in tagcontext
            }
        )
    if expand_tagcontext is True and extract_from_tagcontext is not None:
        df = utils.expand_dataframe_column(df, "tagContext", add_prefix=False)

    if extract_from_staticdataset is not None:
        df["staticDataSet"] = df["staticDataSet"].map(
            lambda staticdataset: [
                data[extract_from_staticdataset] for data in staticdataset
            ]
        )

    if extract_from_pythonmodels is not None and extract_from_pythonmodels in [
        "id",
        "name",
    ]:
        df["pythonModels"] = df["pythonModels"].map(
            lambda pythonModels: [
                data[extract_from_pythonmodels] for data in pythonModels
            ]
        )
    elif extract_from_pythonmodels == "description":
        df["pythonModels"] = df["pythonModels"].map(
            lambda pythonModels: [
                data["pythonModel"]["description"] for data in pythonModels
            ]
        )

    ser = df.squeeze()

    # Output
    return ser


def create_asset_type(
    name: str,
    nature: str,
    tag_context: List[dict],
    dataset: List[dict],
    resources: List[Any],
    static_dataset: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new asset type.

    Parameters
    ----------
    name : str
        The name of the asset type to create.
    nature : {'DEVICE', 'PHYSICAL_ASSET', 'LOGICAL_ASSET'}
        The nature of the asset type to create.
    tag_context : list of dict
        The array of tag values used to label the asset type.
    dataset : list of dict.
        The set of metadata defined at asset type level.
    resources : list of Any.
        The set of resources defined at asset type level.
    static_dataset : list of dict
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the asset type that has just been created.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # endpoint
    response = endpoints.assets.create_asset_type(
        name=name,
        nature=nature,
        tag_context=tag_context,
        dataset=dataset,
        resources=resources,
        static_dataset=static_dataset,
        api_credentials=api_credentials,
    )

    ser = pd.Series(response, dtype=object)

    # Output
    return ser


def get_assets(
    asset_type_id: str,
    tag_value_id: Optional[List[str]] = None,
    query: Optional[int] = None,
    page: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    extract_from_assettype: Optional[str] = "name",
    extract_from_tags: Optional[str] = "value",
    expand_tags: bool = True,
    extract_from_staticdatavalues: Optional[str] = "description",
    expand_staticdatavalues: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Args validation
    if extract_from_assettype not in ["id", "name", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_assettype': {extract_from_assettype}"
        )

    if extract_from_tags not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_tags': {extract_from_tags}"
        )

    if extract_from_tags is None and expand_tags is True:
        raise ValueError("Tags cannot be expanded if 'extract_from_tags' is None")

    if extract_from_staticdatavalues not in ["id", "name", "description", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_staticdatavalues': {extract_from_staticdatavalues}"
        )

    if extract_from_staticdatavalues is None and expand_staticdatavalues is True:
        raise ValueError(
            "Static data values cannot be expanded if 'extract_from_staticdatavalues' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    def get_page(page_num: int):
        page_response = endpoints.assets.get_assets(
            asset_type_id=asset_type_id,
            tag_value_id=tag_value_id,
            query=query,
            page=page_num,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "name", "description"])

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Format dataframe
    if extract_from_assettype is not None:
        df["assetType"] = df["assetType"].map(
            lambda asset_type: asset_type[extract_from_assettype]
        )

    if extract_from_tags == "value":
        df["tags"] = df["tags"].map(
            lambda tags: {tag["tagKey"]["key"]: tag["value"] for tag in tags}
        )

    elif extract_from_tags == "id":
        df["tags"] = df["tags"].map(
            lambda tags: {tag["tagKey"]["id"]: tag["id"] for tag in tags}
        )

    if expand_tags is True and extract_from_tags is not None:
        df = utils.expand_dataframe_column(df, "tags", add_prefix=False)

    if extract_from_staticdatavalues is not None:
        df["staticDataValues"] = df["staticDataValues"].map(
            lambda sdv: {
                sd["assetTypeStaticData"][extract_from_staticdatavalues]: sd["value"]
                for sd in sdv
            }
        )

    if expand_staticdatavalues is True and extract_from_staticdatavalues is not None:
        df = utils.expand_dataframe_column(df, "staticDataValues", add_prefix=False)

    # Output
    return df


def get_single_asset(
    asset_id: str,
    extract_from_assettype: Optional[Literal["id", "name"]] = "name",
    extract_from_tags: Optional[Literal["id", "value"]] = "value",
    expand_tags: bool = True,
    extract_from_staticdatavalues: Optional[
        Literal["id", "name", "description"]
    ] = "description",
    expand_staticdatavalues: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Retrieve the details of a specific asset with ots ID.

    Parameters
    ----------
    asset_id : str
        The ID of the asset to retrieve.
    extract_from_assettype : {'id', 'name'}, optional, default 'name'
        Key to be associated to the column 'assetType'.
    extract_from_tags : {'id', 'value'}, optional, default 'value'
        Key to be associated to the column 'tags'.
    expand_tags : bool, default True
        Whether to expand the column tags.
    extract_from_staticdatavalues : {'id', 'name', 'description'}, optional, default 'description'
        Key to be associated to the column 'staticDataValues'.
    expand_staticdatavalues : bool, default True
        Whether to expand the column 'staticDataValues'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas Series containing the details of a specific asset type.
    """

    # Args validation
    if extract_from_tags is None and expand_tags is True:
        raise ValueError("Tags cannot be expanded if 'extract_from_tags' is None")

    if extract_from_staticdatavalues is None and expand_staticdatavalues is True:
        raise ValueError(
            "Static data values cannot be expanded if 'extract_from_staticdatavalues' is None"
        )

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.assets.get_single_asset(
        asset_id=asset_id, api_credentials=api_credentials
    )

    df = pd.DataFrame([response])
    df.set_index("id", inplace=True)

    # Format dataframe
    if extract_from_assettype is not None:
        df["assetType"] = df["assetType"].map(
            lambda asset_type: asset_type[extract_from_assettype]
        )

    if extract_from_tags == "value":
        df["tags"] = df["tags"].map(
            lambda tags: {tag["tagKey"]["key"]: tag["value"] for tag in tags}
        )

    elif extract_from_tags == "id":
        df["tags"] = df["tags"].map(
            lambda tags: {tag["tagKey"]["id"]: tag["id"] for tag in tags}
        )

    if expand_tags is True and extract_from_tags is not None:
        df = utils.expand_dataframe_column(df, "tags", add_prefix=False)

    if extract_from_staticdatavalues is not None:
        df["staticDataValues"] = df["staticDataValues"].map(
            lambda sdv: {
                sd["assetTypeStaticData"][extract_from_staticdatavalues]: sd["value"]
                for sd in sdv
            }
        )

    if expand_staticdatavalues is True and extract_from_staticdatavalues is not None:
        df = utils.expand_dataframe_column(df, "staticDataValues", add_prefix=False)

    ser = df.squeeze()

    # Output
    return ser


def update_single_asset_tags_and_values(
    asset_id: str,
    tag_values: Optional[Dict] = None,
    static_data_values: Optional[Dict] = None,
    static_data_units: Optional[Dict] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Init
    if tag_values is None:
        tag_values = {}

    if static_data_values is None:
        static_data_values = {}

    if static_data_units is None:
        static_data_units = {}

    # Build payload
    tag_commands = []
    for tagkey_id, tagvalue in tag_values.items():
        tag_commands.append({"tagKeyId": tagkey_id, "id": None, "value": tagvalue})

    value_commands = []
    for data_id, value in static_data_values.items():
        value_commands.append(
            {
                "assetTypeStaticDataId": data_id,
                "value": value,
                "unitId": static_data_units.get(data_id),
            }
        )

    # Query endpoint
    response = endpoints.assets.update_single_asset_tags_and_values(
        asset_id=asset_id,
        tag_commands=tag_commands,
        value_commands=value_commands,
        api_credentials=api_credentials,
    )

    # Output
    return response


def create_asset(
    name: str,
    external_reference: str,
    description: str,
    asset_type_id: str,
    tag_values: List[dict],
    data_mappings: List[dict],
    static_data_values: List[dict],
    file_resource_ids: Optional[List[dict]] = None,
    html_resources: Optional[List[dict]] = None,
    logistic_route_ids: Optional[List[dict]] = None,
    python_model_instances: Optional[List[dict]] = None,
    extract_from_assettype: Optional[Literal["id", "name"]] = "name",
    extract_from_tags: Optional[Literal["id", "value"]] = "value",
    expand_tags: bool = True,
    extract_from_staticdatavalues: Optional[
        Literal["id", "name", "description"]
    ] = "description",
    expand_staticdatavalues: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new asset.

    Parameters
    ----------
    name : str
        The name of the asset to create.
    external_reference : str
        The reference of the asset in external system.
    description : str
        An optional description used to understand what this asset means.
    asset_type_id : str
        The asset type this asset is attached to.
    tag_values : list of dict
        The array of tag values used to label the asset.
        Each dict should contain: 'tagKeyId', 'id', 'value'.
    data_mappings : list of dict
        The data mapping between AssetTypeData and Data for this asset.
        Each dict should contain: 'assetTypeDataId', 'mode', 'dataId' (if mode is MAPPED).
    static_data_values : list of dict
        The static data value of asset static data attached to this asset.
        Each dict should contain: 'assetTypeStaticDataId', 'value', 'unitId' (optional).
    file_resource_ids : list of dict, optional
        The file resources to attach to the asset.
    html_resources : list of dict, optional
        The HTML resources to attach to the asset.
    logistic_route_ids : list of dict, optional
        The logistic routes to attach to the asset.
    python_model_instances : list of dict, optional
        The instances of python models to create or update.
    extract_from_assettype : {'id', 'name'}, optional, default 'name'
        Key to be associated to the column 'assetType'.
    extract_from_tags : {'id', 'value'}, optional, default 'value'
        Key to be associated to the column 'tags'.
    expand_tags : bool, default True
        Whether to expand the column tags.
    extract_from_staticdatavalues : {'id', 'name', 'description'}, optional, default 'description'
        Key to be associated to the column 'staticDataValues'.
    expand_staticdatavalues : bool, default True
        Whether to expand the column 'staticDataValues'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the asset that has just been created.
    """

    # Args validation
    if extract_from_tags is None and expand_tags is True:
        raise ValueError("Tags cannot be expanded if 'extract_from_tags' is None")

    if extract_from_staticdatavalues is None and expand_staticdatavalues is True:
        raise ValueError(
            "Static data values cannot be expanded if 'extract_from_staticdatavalues' is None"
        )

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.assets.create_asset(
        name=name,
        external_reference=external_reference,
        description=description,
        asset_type_id=asset_type_id,
        tag_values=tag_values,
        data_mappings=data_mappings,
        static_data_values=static_data_values,
        file_resource_ids=file_resource_ids,
        html_resources=html_resources,
        logistic_route_ids=logistic_route_ids,
        python_model_instances=python_model_instances,
        api_credentials=api_credentials,
    )

    df = pd.DataFrame([response])
    df.set_index("id", inplace=True)

    # Format dataframe
    if extract_from_assettype is not None:
        df["assetType"] = df["assetType"].map(
            lambda asset_type: asset_type[extract_from_assettype]
        )

    if extract_from_tags == "value":
        df["tags"] = df["tags"].map(
            lambda tags: {tag["tagKey"]["key"]: tag["value"] for tag in tags}
        )

    elif extract_from_tags == "id":
        df["tags"] = df["tags"].map(
            lambda tags: {tag["tagKey"]["id"]: tag["id"] for tag in tags}
        )

    if expand_tags is True and extract_from_tags is not None:
        df = utils.expand_dataframe_column(df, "tags", add_prefix=False)

    if extract_from_staticdatavalues is not None:
        df["staticDataValues"] = df["staticDataValues"].map(
            lambda sdv: {
                sd["assetTypeStaticData"][extract_from_staticdatavalues]: sd["value"]
                for sd in sdv
            }
        )

    if expand_staticdatavalues is True and extract_from_staticdatavalues is not None:
        df = utils.expand_dataframe_column(df, "staticDataValues", add_prefix=False)

    ser = df.squeeze()

    # Output
    return ser



from .single_observation import get_single_obs_time_data, get_single_obs_batch_data
from .files import get_resource_file, update_instance_resource
from .instances import invoke_instance

__all__ = [
    "get_single_obs_time_data",
    "get_single_obs_batch_data",
    "get_resource_file",
    "update_instance_resource",
    "invoke_instance",
]
